package dev.shadowsoffire.apotheosis.ench;

import java.math.BigDecimal;
import s_com.udojava.evalex.Expression;
import dev.shadowsoffire.apotheosis.ench.table.EnchantingStatRegistry;
import net.minecraftforge.registries.ForgeRegistries;
import dev.shadowsoffire.placebo.config.Configuration;
import net.minecraft.world.item.enchantment.Enchantment;

public class EnchantmentInfo
{
    protected final Enchantment ench;
    protected final int maxLevel;
    protected final int maxLootLevel;
    protected final boolean treasure;
    protected final boolean discoverable;
    protected final boolean lootable;
    protected final boolean tradeable;
    protected final PowerFunc maxPower;
    protected final PowerFunc minPower;
    
    public EnchantmentInfo(final Enchantment ench, final int maxLevel, final int maxLootLevel, final PowerFunc max, final PowerFunc min, final boolean treasure, final boolean discoverable, final boolean lootable, final boolean tradeable) {
        this.ench = ench;
        this.maxLevel = maxLevel;
        this.maxLootLevel = maxLootLevel;
        this.maxPower = max;
        this.minPower = min;
        this.treasure = treasure;
        this.discoverable = discoverable;
        this.lootable = lootable;
        this.tradeable = tradeable;
    }
    
    @Deprecated
    public EnchantmentInfo(final Enchantment ench) {
        this(ench, ench.m_6586_(), ench.m_6586_(), defaultMax(ench), defaultMin(ench), ench.m_6591_(), ench.m_6592_(), ench.m_6592_(), ench.m_6594_());
    }
    
    public int getMaxLevel() {
        return Math.min(EnchModule.ENCH_HARD_CAPS.getOrDefault((Object)this.ench, 127), this.maxLevel);
    }
    
    public int getMaxLootLevel() {
        return Math.min(EnchModule.ENCH_HARD_CAPS.getOrDefault((Object)this.ench, 127), this.maxLootLevel);
    }
    
    public int getMinPower(final int level) {
        return this.minPower.getPower(level);
    }
    
    public int getMaxPower(final int level) {
        return this.maxPower.getPower(level);
    }
    
    public boolean isDiscoverable() {
        return this.discoverable;
    }
    
    public boolean isTreasure() {
        return this.treasure;
    }
    
    public boolean isLootable() {
        return this.lootable;
    }
    
    public boolean isTradeable() {
        return this.tradeable;
    }
    
    public static EnchantmentInfo load(final Enchantment ench, final Configuration cfg) {
        final String category = ForgeRegistries.ENCHANTMENTS.getKey((Object)ench).toString();
        final int max = cfg.getInt("Max Level", category, EnchModule.getDefaultMax(ench), 1, 127, "The max level of this enchantment - originally " + ench.m_6586_());
        final int maxLoot = cfg.getInt("Max Loot Level", category, ench.m_6586_(), 1, 127, "The max level of this enchantment available from loot sources.");
        final String maxF = cfg.getString("Max Power Function", category, "", "A function to determine the max enchanting power.  The variable \"x\" is level.  See: https://github.com/uklimaschewski/EvalEx#usage-examples");
        final String minF = cfg.getString("Min Power Function", category, "", "A function to determine the min enchanting power.");
        final PowerFunc maxPower = maxF.isEmpty() ? defaultMax(ench) : new ExpressionPowerFunc(maxF);
        final PowerFunc minPower = minF.isEmpty() ? defaultMin(ench) : new ExpressionPowerFunc(minF);
        final boolean treasure = cfg.getBoolean("Treasure", category, ench.m_6591_(), "If this enchantment is only available by loot sources.");
        final boolean discoverable = cfg.getBoolean("Discoverable", category, ench.m_6592_(), "If this enchantment is obtainable via enchanting and enchanted loot items.");
        final boolean lootable = cfg.getBoolean("Lootable", category, ench.m_6592_(), "If enchanted books of this enchantment are available via loot sources.");
        final boolean tradeable = cfg.getBoolean("Tradeable", category, ench.m_6594_(), "If enchanted books of this enchantment are available via villager trades.");
        final EnchantmentInfo info = new EnchantmentInfo(ench, max, maxLoot, maxPower, minPower, treasure, discoverable, lootable, tradeable);
        final String rarity = cfg.getString("Rarity", category, ench.m_44699_().name(), "The rarity of this enchantment.  Valid values are COMMON, UNCOMMON, RARE, and VERY_RARE.");
        try {
            final Enchantment.Rarity r = Enchantment.Rarity.valueOf(rarity);
            ench.f_44674_ = r;
        }
        catch (final Exception ex) {
            EnchModule.LOGGER.error("Failed to parse rarity for {}, as {} is not a valid rarity string.", (Object)category, (Object)rarity);
        }
        return info;
    }
    
    public static PowerFunc defaultMax(final Enchantment ench) {
        return level -> (int)(EnchantingStatRegistry.getAbsoluteMaxEterna() * 4.0f);
    }
    
    public static PowerFunc defaultMin(final Enchantment ench) {
        return level -> {
            if (level > ench.m_6586_() && level > 1) {
                int diff = ench.m_6183_(ench.m_6586_()) - ench.m_6183_(ench.m_6586_() - 1);
                if (diff == 0) {
                    diff = 15;
                }
                return ench.m_6183_(level) + diff * (int)Math.pow(level - ench.m_6586_(), 1.6);
            }
            else {
                return ench.m_6183_(level);
            }
        };
    }
    
    public static class ExpressionPowerFunc implements PowerFunc
    {
        Expression ex;
        
        public ExpressionPowerFunc(final String func) {
            this.ex = new Expression(func);
        }
        
        @Override
        public int getPower(final int level) {
            return this.ex.setVariable("x", new BigDecimal(level)).eval().intValue();
        }
    }
    
    public interface PowerFunc
    {
        int getPower(final int p0);
    }
}
